import configparser
import shutil
import logging
import subprocess
import time
import os
import sys
from . import PluginInstall, PluginType

from meltano.core.error import SubprocessError
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.config_service import PluginConfigService
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import invoker_factory, PluginInvoker
from meltano.core.db import project_engine
from meltano.core.venv_service import VenvService
from meltano.core.utils import nest, map_dict


class AirflowInvoker(PluginInvoker):
    def env(self):
        env = super().env()
        venv_dir = self.project.venvs_dir(self.plugin.type, self.plugin.name)

        # add the Airflow virtualenv because it contains `gunicorn`
        env["PATH"] = os.pathsep.join([str(venv_dir.joinpath("bin")), env["PATH"]])
        env["VIRTUAL_ENV"] = str(venv_dir)
        env["AIRFLOW_HOME"] = str(self.config_service.run_dir)

        # we want Airflow to inherit our current venv so that the `meltano`
        # module can be used from inside `orchestrate/dags/meltano.py`
        sys_paths = []
        for path in sys.path:
            # the current venv if any
            if path.endswith("site-packages"):
                sys_paths.append(path)

            # when meltano is installed as editable
            if path.endswith(os.path.join("meltano", "src")):
                sys_paths.append(path)

        if "PYTHONPATH" in env and env["PYTHONPATH"]:
            sys_paths.append(env["PYTHONPATH"])

        env["PYTHONPATH"] = os.pathsep.join(sys_paths)

        return env


class Airflow(PluginInstall):
    __plugin_type__ = PluginType.ORCHESTRATORS
    __invoker_cls__ = AirflowInvoker

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @property
    def config_files(self):
        return {"config": "airflow.cfg"}

    @hook("before_install")
    def setup_env(self, project, args=[]):
        # to make airflow installables without GPL dependency
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"

    @hook("before_configure")
    def before_configure(self, invoker, session):
        project = invoker.project

        stub_path = project.plugin_dir(self).joinpath("airflow.cfg")

        # generate the default `airflow.cfg`
        handle = invoker.invoke(
            "--help",
            require_preparation=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        handle.wait()

        airflow_cfg_path = invoker.files["config"]
        logging.debug(f"Generated default '{str(airflow_cfg_path)}'")

        # move it to the config dir
        shutil.move(airflow_cfg_path, stub_path)
        airflow_cfg_path = stub_path
        logging.debug(f"Moved to '{str(stub_path)}'")

        # open the configuration and update it
        # now we let's update the config to use our stubs
        airflow_cfg = configparser.ConfigParser()

        with airflow_cfg_path.open() as cfg:
            airflow_cfg.read_file(cfg)
            logging.debug(f"Loaded '{str(airflow_cfg_path)}'")

        config = {}
        for key, value in invoker.plugin_config.items():
            nest(config, key, str(value))

        for section, cfg in config.items():
            airflow_cfg[section].update(cfg)
            logging.debug(f"\tUpdated section [{section}] with {cfg}")

        with airflow_cfg_path.open("w") as cfg:
            airflow_cfg.write(cfg)
            logging.debug(f"Saved '{str(airflow_cfg_path)}'")

        # we've changed the configuration here, so we need to call
        # prepare again on the invoker so it re-reads the configuration
        # for the Airflow plugin
        invoker.prepare(session)
        handle = invoker.invoke(
            "initdb",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        initdb = handle.wait()

        if initdb:
            raise SubprocessError("airflow initdb failed", handle)

        logging.debug(f"Completed `airflow initdb`")



import os
import logging
import warnings
import meltano
import importlib


workers = int(os.getenv("WORKERS", 4))
timeout = 600


def on_reload(arbiter):
    importlib.reload(meltano)
    logging.info(f"Meltano version reloaded to {meltano.__version__}")



import os
import sqlalchemy
import logging
from typing import Iterable, Dict, Tuple, List
from copy import deepcopy
from enum import Enum
import re

from meltano.core.utils import nest, find_named, setting_env, NotFound, truthy
from meltano.core.config_service import ConfigService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.error import Error
from . import PluginRef, PluginType, Plugin, PluginInstall, SettingDefinition, Profile
from .error import PluginMissingError
from .setting import PluginSetting, REDACTED_VALUE


class PluginSettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, plugin: PluginRef, name: str):
        super().__init__(f"Cannot find setting {name} in {plugin}")


class PluginSettingValueSource(str, Enum):
    CONFIG_OVERRIDE = "config_override"  # 0
    ENV = "env"  # 1
    MELTANO_YML = "meltano_yml"  # 2
    DB = "db"  # 3
    DEFAULT = "default"  # 4


class PluginSettingsService:
    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
        show_hidden=True,
        env_override={},
        config_override={},
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.show_hidden = show_hidden
        self.env_override = env_override
        self.config_override = config_override

        self._env = None

    @classmethod
    def unredact(cls, values: dict) -> Dict:
        """
        Removes any redacted values in a dictionary.
        """

        return {k: v for k, v in values.items() if v != REDACTED_VALUE}

    @classmethod
    def is_kind_redacted(cls, kind) -> bool:
        return kind in ("password", "oauth")

    def with_env_override(self, env_override):
        return self.__class__(
            self.project,
            self.config_service,
            self.discovery_service,
            self.show_hidden,
            {**self.env_override, **env_override},
            self.config_override,
        )

    def with_config_override(self, config_override):
        return self.__class__(
            self.project,
            self.config_service,
            self.discovery_service,
            self.show_hidden,
            self.env_override,
            {**self.config_override, **config_override},
        )

    @property
    def env(self):
        if not self._env:
            self._env = {**os.environ, **self.env_override}

        return self._env

    def profile_with_config(
        self, session, plugin: PluginRef, profile: Profile, redacted=False
    ):
        plugin_install = self.get_install(plugin)

        plugin_install.use_profile(profile)

        full_config = self.config_with_sources(
            session, plugin_install, redacted=redacted
        )

        return {
            **profile.canonical(),
            "config": {key: config["value"] for key, config in full_config.items()},
            "config_sources": {
                key: config["source"] for key, config in full_config.items()
            },
        }

    def profiles_with_config(
        self, session, plugin: PluginRef, redacted=False
    ) -> List[Dict]:
        plugin_install = self.get_install(plugin)

        return [
            self.profile_with_config(session, plugin, profile, redacted=redacted)
            for profile in (Profile.DEFAULT, *plugin_install.profiles)
        ]

    def config_with_sources(
        self,
        session,
        plugin: PluginRef,
        sources: List[PluginSettingValueSource] = None,
        redacted=False,
    ):
        setting_names = [
            *(setting.name for setting in self.definitions(plugin)),
            *(key for key in self.get_install(plugin).current_config.keys()),
        ]

        config = {}
        for name in setting_names:
            value, source = self.get_value(session, plugin, name, redacted=redacted)
            if sources and source not in sources:
                continue

            config[name] = {"value": value, "source": source}

        return config

    def as_config(self, *args, **kwargs) -> Dict:
        full_config = self.config_with_sources(*args, **kwargs)

        return {key: config["value"] for key, config in full_config.items()}

    def as_env(
        self, session, plugin: PluginRef, sources: List[PluginSettingValueSource] = None
    ) -> Dict[str, str]:
        # defaults to the meltano.yml for extraneous settings
        plugin_def = self.get_definition(plugin)
        env = {}

        for setting in self.definitions(plugin):
            value, source = self.get_value(session, plugin, setting.name)
            if sources and source not in sources:
                logging.debug(f"Setting {setting.name} is not in sources: {sources}.")
                continue

            if value is None:
                continue

            env_key = self.setting_env(setting, plugin_def)
            env[env_key] = str(value)

        return env

    def set(self, session, plugin: PluginRef, name: str, value, enabled=True):
        try:
            plugin_def = self.get_definition(plugin)
            setting_def = self.find_setting(plugin, name)
            env_key = self.setting_env(setting_def, plugin_def)

            if env_key in self.env:
                logging.warning(f"Setting `{name}` is currently set via ${env_key}.")
                return

            if value == REDACTED_VALUE:
                return

            setting = PluginSetting(
                namespace=plugin.qualified_name, name=name, value=value, enabled=enabled
            )

            session.merge(setting)
            session.commit()

            return setting
        except StopIteration:
            logging.warning(f"Setting `{name}` not found.")

    def unset(self, session, plugin: PluginRef, name: str):
        session.query(PluginSetting).filter_by(
            namespace=plugin.qualified_name, name=name
        ).delete()
        session.commit()

    def get_definition(self, plugin: PluginRef) -> Plugin:
        return self.discovery_service.find_plugin(plugin.type, plugin.name)

    def find_setting(self, plugin: PluginRef, name: str) -> Dict:
        try:
            return find_named(self.definitions(plugin), name)
        except NotFound as err:
            raise PluginSettingMissingError(plugin, name) from err

    def definitions(self, plugin_ref: PluginRef) -> Iterable[Dict]:
        settings = set()
        plugin_def = self.get_definition(plugin_ref)
        only_visible = lambda s: s.kind != "hidden" or self.show_hidden

        return list(filter(only_visible, plugin_def.settings))

    def get_install(self, plugin: PluginRef) -> PluginInstall:
        return self.config_service.get_plugin(plugin)

    def setting_env(self, setting_def, plugin_def):
        return setting_def.env or setting_env(plugin_def.namespace, setting_def.name)

    def cast_value(self, setting_def, value):
        if isinstance(value, str) and setting_def.kind == "boolean":
            value = truthy(value)

        return value

    def get_value(self, session, plugin: PluginRef, name: str, redacted=False):
        plugin_install = self.get_install(plugin)
        plugin_def = self.get_definition(plugin)

        try:
            setting_def = self.find_setting(plugin, name)
        except PluginSettingMissingError:
            setting_def = None

        def config_override_getter():
            try:
                return self.config_override[name]
            except KeyError:
                return None

        def env_getter():
            if not setting_def:
                return None

            env_key = self.setting_env(setting_def, plugin_def)

            try:
                return self.env[env_key]
            except KeyError:
                return None
            else:
                logging.debug(
                    f"Found ENV variable {env_key} for {plugin_def.namespace}:{name}"
                )

        def meltano_yml_getter():
            try:
                value = plugin_install.current_config[name]
                return self.expand_env_vars(value)
            except KeyError:
                return None

        def db_getter():
            try:
                return (
                    session.query(PluginSetting)
                    .filter_by(namespace=plugin.qualified_name, name=name, enabled=True)
                    .one()
                    .value
                )
            except sqlalchemy.orm.exc.NoResultFound:
                return None

        def default_getter():
            if not setting_def:
                return None
            return self.expand_env_vars(setting_def.value)

        config_getters = {
            PluginSettingValueSource.CONFIG_OVERRIDE: config_override_getter,
            PluginSettingValueSource.ENV: env_getter,
            PluginSettingValueSource.MELTANO_YML: meltano_yml_getter,
            PluginSettingValueSource.DB: db_getter,
            PluginSettingValueSource.DEFAULT: default_getter,
        }

        for source, getter in config_getters.items():
            value = getter()
            if value is not None:
                break

        if setting_def:
            value = self.cast_value(setting_def, value)

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and self.is_kind_redacted(setting_def.kind):
                value = REDACTED_VALUE

        return value, source

    def expand_env_vars(self, raw_value):
        if not isinstance(raw_value, str):
            return raw_value

        # find viable substitutions
        var_matcher = re.compile(
            """
            \$                 # starts with a '$'
            (?:                # either $VAR or ${VAR}
                {(\w+)}|(\w+)  # capture the variable name as group[0] or group[1]
            )
            """,
            re.VERBOSE,
        )

        def subst(match) -> str:
            try:
                # the variable can be in either group
                var = next(var for var in match.groups() if var)
                val = str(self.env[var])

                if not val:
                    logging.warning(f"Variable {var} is empty.")

                return val
            except KeyError as e:
                logging.warning(f"Variable {var} is missing from the environment.")
                return None

        fullmatch = re.fullmatch(var_matcher, raw_value)
        if fullmatch:
            # If the entire value is an env var reference, return None if it isn't set
            return subst(fullmatch)

        return re.sub(var_matcher, subst, raw_value)



from typing import Optional
from collections import namedtuple

from meltano.core.project import Project
from meltano.core.job import Job
from meltano.core.config_service import ConfigService
from meltano.core.plugin import Plugin, PluginType, PluginRef
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.utils import flatten


class PluginContext(
    namedtuple("PluginContext", "ref install definition settings_service session")
):
    @property
    def namespace(self):
        return self.definition.namespace

    @property
    def name(self):
        return self.install.name

    @property
    def type(self):
        return self.ref.type

    @property
    def config(self):
        return self.settings_service.as_config(self.session, self.ref)

    @property
    def config_env(self):
        return self.settings_service.as_env(self.session, self.ref)


class ELTContext:
    def __init__(
        self,
        project,
        session,
        job: Optional[Job] = None,
        extractor: Optional[PluginContext] = None,
        loader: Optional[PluginContext] = None,
        transformer: Optional[PluginContext] = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.job = job
        self.extractor = extractor
        self.loader = loader
        self.transformer = transformer
        self.plugin_discovery_service = (
            plugin_discovery_service or PluginDiscoveryService(project)
        )

    @property
    def elt_run_dir(self):
        if self.job:
            self.project.job_dir(self.job.job_id, str(self.job.run_id))

    def extractor_invoker(self):
        return invoker_factory(
            self.project,
            self.extractor.install,
            run_dir=self.elt_run_dir,
            plugin_settings_service=self.extractor.settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )

    def loader_invoker(self):
        return invoker_factory(
            self.project,
            self.loader.install,
            run_dir=self.elt_run_dir,
            plugin_settings_service=self.loader.settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )

    def transformer_invoker(self):
        return invoker_factory(
            self.project,
            self.transformer.install,
            plugin_settings_service=self.transformer.settings_service,
            plugin_discovery_service=self.plugin_discovery_service,
        )


class ELTContextBuilder:
    def __init__(
        self,
        project: Project,
        config_service: ConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)
        self.plugin_discovery_service = (
            plugin_discovery_service
            or PluginDiscoveryService(project, config_service=config_service)
        )
        self.plugin_settings_service = plugin_settings_service or PluginSettingsService(
            project,
            config_service=config_service,
            plugin_discovery_service=plugin_discovery_service,
        )
        self._extractor = None
        self._loader = None
        self._transformer = None
        self._job = None

    def with_extractor(self, extractor_name: str):
        self._extractor = PluginRef(PluginType.EXTRACTORS, extractor_name)

        return self

    def with_loader(self, loader_name: str):
        self._loader = PluginRef(PluginType.LOADERS, loader_name)

        return self

    def with_transform(self, transform: str):
        if transform == "skip":
            return self

        return self.with_transformer("dbt")

    def with_transformer(self, transformer_name: str):
        self._transformer = PluginRef(PluginType.TRANSFORMERS, transformer_name)

        return self

    def with_job(self, job: Job):
        self._job = job

        return self

    def plugin_context(self, session, plugin: PluginRef, env={}):
        return PluginContext(
            ref=plugin,
            install=self.config_service.find_plugin(
                plugin_name=plugin.name, plugin_type=plugin.type
            ),
            definition=self.plugin_discovery_service.find_plugin(
                plugin_name=plugin.name, plugin_type=plugin.type
            ),
            settings_service=self.plugin_settings_service.with_env_override(env),
            session=session,
        )

    def context(self, session) -> ELTContext:
        env = {}

        def env_for_plugin(plugin):
            env_struct = {
                "meltano": {
                    plugin.type.singular: {  # MELTANO_EXTRACTOR_...
                        "name": plugin.name,
                        "namespace": plugin.namespace,
                    },
                    plugin.type.singular[0:-2]: plugin.config,  # MELTANO_EXTRACT_...
                },
                **plugin.config_env,  # TAP_...
            }
            env_vars = flatten(env_struct, "env_var").items()

            return {k: str(v) for k, v in env_vars if v is not None}

        extractor = None
        if self._extractor:
            extractor = self.plugin_context(session, self._extractor, env)
            env.update(env_for_plugin(extractor))

        loader = None
        if self._loader:
            loader = self.plugin_context(session, self._loader, env)
            env.update(env_for_plugin(loader))

        transformer = None
        if self._transformer:
            transformer = self.plugin_context(session, self._transformer, env)
            env.update(env_for_plugin(transformer))

        return ELTContext(
            self.project,
            session,
            job=self._job,
            extractor=extractor,
            loader=loader,
            transformer=transformer,
            plugin_discovery_service=self.plugin_discovery_service,
        )



"""add last activity at column

Revision ID: e4fbabc3fed6
Revises: 367228df6a43
Create Date: 2020-03-24 15:53:54.142685

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e4fbabc3fed6"
down_revision = "367228df6a43"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("user", sa.Column("last_activity_at", sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column("user", "last_activity_at")



import os
import asyncio
import logging
import sys
from io import StringIO

from . import Runner
from meltano.core.error import SubprocessError
from meltano.core.project import Project
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.db import project_engine
from meltano.core.logging import capture_subprocess_output
from meltano.core.elt_context import ELTContext


class DbtRunner(Runner):
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

    @property
    def project(self):
        return self.context.project

    @property
    def plugin(self):
        return self.context.transformer

    async def invoke(self, dbt: PluginInvoker, cmd, *args, **kwargs):
        handle = await dbt.invoke_async(
            cmd,
            *args,
            **kwargs,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        await asyncio.wait(
            [
                capture_subprocess_output(handle.stdout, sys.stdout),
                capture_subprocess_output(handle.stderr, sys.stderr),
                handle.wait(),
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

        if handle.returncode:
            raise Exception(
                f"dbt {cmd} didn't exit cleanly. Exit code: {handle.returncode}"
            )

    def run(self, session, dry_run=False):
        dbt = self.context.transformer_invoker()
        dbt.prepare(session)

        # Get an asyncio event loop and use it to run the dbt commands
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.invoke(dbt, "clean"))
        loop.run_until_complete(self.invoke(dbt, "deps"))

        cmd = "compile" if dry_run else "run"
        loop.run_until_complete(
            self.invoke(dbt, cmd, "--models", str(self.plugin.config["models"]))
        )
